from asyncio.coroutines import iscoroutinefunction
from asyncio.events import AbstractEventLoop, get_running_loop
from concurrent.futures import ThreadPoolExecutor
from functools import partial

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable

loop: AbstractEventLoop


async def wrap_async[**P, T](
    func: "Callable[P, T]",
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    if iscoroutinefunction(func) or not callable(func):
        err = f"{func} is either a coroutine or not callable"
        raise TypeError(err)
    global loop
    try:
        _ = loop
    except NameError:
        loop = get_running_loop()
    with ThreadPoolExecutor(1) as e:
        return await loop.run_in_executor(e, partial(func, *args, **kwargs))



from datetime import timedelta
from pathlib import Path

from msgspec import json

from tg_util.src import DefaultARG

TYPE_CHECKING = False
if TYPE_CHECKING:
    from argparse import ArgumentParser
    from typing import Any
    from urllib.parse import ParseResult

JSON_ENC = json.Encoder()


def round_size(n: float | int) -> str:
    for unit in ("Bytes", "KB", "MB"):
        if n < 1024.0:
            if unit == "Bytes":
                return f"{n} {unit}"
            else:
                return f"{n:.1f} {unit}"
        n /= 1024.0
    return f"{n:.1f} GB"


def format_duration(secs: float) -> str:
    return str(timedelta(seconds=secs))


def parse_proxy(url: "ParseResult"):
    from python_socks import ProxyType

    proxy: dict[str, Any] = {}
    match url.scheme.lower():
        case "socks" | "socks5":
            proxy["proxy_type"] = ProxyType.SOCKS5
        case "socks4":
            proxy["proxy_type"] = ProxyType.SOCKS4
        case "http" | "https":
            proxy["proxy_type"] = ProxyType.HTTP
    proxy["addr"] = url.hostname
    proxy["port"] = url.port
    if url.username:
        proxy["username"] = url.username
    if url.password:
        proxy["password"] = url.password
    return proxy


def encode_json_str(data: "Any"):
    return JSON_ENC.encode(data).decode()


def add_misc_args(parser: "ArgumentParser", version: str):
    misc = parser.add_argument_group("misc")
    misc.add_argument("-h", "--help", action="help", help="print this help and exit")
    misc.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=DefaultARG(value=False),
        help="enable debug log",
        dest="debug",
    )
    misc.add_argument(
        "-V",
        "--version",
        action="version",
        help="print version",
        version=f"%(prog)s {version}",
    )
    return misc


def add_opts_args(parser: "ArgumentParser"):
    options = parser.add_argument_group("options")
    options.add_argument(
        "-a",
        "--archive",
        dest="archive",
        default=DefaultARG("sqlite::memory:"),
        metavar="scheme:[//user:pass@host:port/]schema",
    )
    options.add_argument(
        "-s",
        "--session",
        dest="session",
        default=DefaultARG("sqlite:telethon.session"),
        metavar="scheme:[//user:pass@host:port/]schema[?api_id=id&api_hash=hash]",
    )
    options.add_argument(
        "-c",
        "--config",
        type=Path,
        default=None,
        help="load config from FILE",
        dest="config",
        metavar="FILE",
    )
    options.add_argument(
        "--proxy",
        default=DefaultARG(None),
        dest="proxy",
        metavar="{http,socks4,socks5}://user:pass@host:port",
    )
    return options


def unpack_default[RT, DT](arg: DefaultARG[DT] | RT) -> DT | RT:
    if isinstance(arg, DefaultARG):
        return arg.value
    return arg



from asyncio import Lock
from datetime import datetime

from asyncpg import Connection, Record, UniqueViolationError, connect

from .base import ArchiveBase

TYPE_CHECKING = False
if TYPE_CHECKING:
    from urllib.parse import ParseResult

    from _typeshed import Unused

    from tg_util.src.tg.messages.export import MessageExport


class PSQLArchive(ArchiveBase):
    _conn: "Connection[Record]"

    def __init__(self, params: "ParseResult"):
        self._params = params
        self._lock = Lock()

    async def __aenter__(self):
        self._conn = await connect(self._params.geturl())
        return self

    async def __aexit__(self, *_exc: "Unused"):
        await self._conn.close()

    async def prepare(self):
        q = (
            """CREATE TABLE IF NOT EXISTS _archive_ (
                file_id int8 NOT NULL,
                msg text NOT NULL,
                msg_id int4 NOT NULL,
                chat_id int8 NOT NULL,
                username text NULL,
                hash bytea NOT NULL,
                width int4 NULL,
                height int4 NULL,
                "size" int8 NULL,
                duration float8 NULL,
                downloaded timestamp NULL,
                "type" varchar(20) NOT NULL,
                CONSTRAINT _archive__pk PRIMARY KEY (file_id),
                CONSTRAINT _archive__unique UNIQUE (hash)
            );"""
            'CREATE INDEX IF NOT EXISTS "_archive__type_IDX" ON "_archive_" ("type");'
        )
        await self._conn.execute(q)

    async def check_attr(
        self,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
    ):
        async with self._lock:
            r = await self._conn.fetchrow(
                "select msg, hash, downloaded from _archive_ where "
                "downloaded is not null and (hash = $1 or "
                "(width = $2 and height = $3 and size = $4 and duration = $5))",
                hash,
                width,
                height,
                size,
                duration,
            )
        if r:
            return r[0], r[1], r[2]

    async def check_id(self, file_id: int):
        async with self._lock:
            return await self._conn.fetchval(
                "select msg from _archive_ where file_id = $1 and downloaded is not null",
                file_id,
            )

    async def set_complete(self, file_id: int):
        async with self._lock:
            await self._conn.execute(
                "update _archive_ set downloaded = $1 where file_id = $2",
                datetime.now(),
                file_id,
            )

    async def update(
        self,
        file_id: int,
        msg: str,
        msg_id: int,
        chat_id: int,
        chat_username: str | None,
        hash: bytes,
        width: int | None,
        height: int | None,
        size: int | None,
        duration: float | None,
        type: str,
    ):
        insert = (
            "insert into _archive_ (file_id, msg, msg_id, chat_id, "
            "username, hash, width, height, size, duration, type) "
            "values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11)",
            file_id,
            msg,
            msg_id,
            chat_id,
            chat_username,
            hash,
            width,
            height,
            size,
            duration,
            type,
        )
        async with self._lock:
            try:
                await self._conn.execute(*insert)
            except UniqueViolationError:
                async with self._conn.transaction():
                    await self._conn.execute(
                        "delete from _archive_ where file_id = $1 or hash = $2",
                        file_id,
                        hash,
                    )
                    await self._conn.execute(*insert)

    async def export(self, message: "MessageExport"):
        insert = (
            "insert into _all_chats_ values ($1, $2, $3, $4, $5, $6, $7, "
            "$8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, "
            "$20, $21, $22, $23, $24, $25, $26, $27, $28)",
            *message.as_tuple(),
        )
        try:
            await self._conn.execute(*insert)
        except UniqueViolationError:
            async with self._conn.transaction():
                await self._conn.execute(
                    "delete from _all_chats_ where chat_id = $1 and message_id = $2",
                    message.chat_id,
                    message.message_id,
                )
                await self._conn.execute(*insert)



__all__ = (
    "ABC",
    "ABCMeta",
    "ARGSBase",
    "abstractmethod",
)

from abc import ABCMeta as _ABCMeta
from abc import abstractmethod
from itertools import chain

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any


class ABCMeta(_ABCMeta):
    __repr_fields__: tuple[str, ...]
    __slots__: tuple[str, ...]

    def __new__(
        mcls,
        name: str,
        bases: tuple[type, ...],
        namespace: "dict[str, Any]",
        /,
        **kwargs: "Any",
    ):
        if "__slots__" in namespace:
            err = "__slots__ should not be defined"
            raise TypeError(err)

        slots = ()
        if "__annotations__" in namespace:
            slots = tuple(x for x in namespace["__annotations__"] if x not in namespace)
        namespace["__slots__"] = slots

        fields: list[str] = [
            f for b in bases for f in getattr(b, "__repr_fields__", ())
        ]
        if fields:
            namespace["__repr_fields__"] = tuple(dict.fromkeys(chain(fields, slots)))
        else:
            namespace["__repr_fields__"] = slots
        return super().__new__(mcls, name, bases, namespace, **kwargs)


class ABC(metaclass=ABCMeta):
    pass


class ARGSBase(ABC):
    def __iter_fields__(self):
        for f in sorted(self.__repr_fields__):
            try:
                yield f, getattr(self, f)
            except AttributeError:
                continue

    def __repr__(self) -> str:
        attr = ", ".join(["%s=%r" % f for f in self.__iter_fields__()])
        return f"{self.__class__.__name__}({attr})"


class DefaultARG[T](ABC):
    value: T

    def __init__(self, value: T):
        self.value = value

    def __bool__(self):
        return bool(self.value)

    def __repr__(self):
        return f"<Default: {self.value!r}>"



from pathlib import Path
from typing import ClassVar, Self

from msgspec import Struct, json, yaml

from .enums import FileType

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any


class Decodable(Struct):
    __jdec__: ClassVar[json.Decoder[Self]]

    @classmethod
    def decode_json(cls, buf: bytes, /) -> Self:
        try:
            return cls.__jdec__.decode(buf)
        except AttributeError:
            cls.__jdec__ = json.Decoder(cls, dec_hook=dec_hook)
            return cls.decode_json(buf)

    @classmethod
    def decode_yaml(cls, buf: bytes, /) -> Self:
        return yaml.decode(buf, type=cls, dec_hook=dec_hook)


class TLSchemaBase(Struct, tag=True, tag_field="_"):
    pass


class FileAttribute(Struct):
    width: int | None
    height: int | None
    duration: float | None
    size: int | None
    type: FileType
    id: int


class EntityStats(Struct):
    type: str
    title: str
    username: str
    id: int
    medias: int
    files: int
    messages: int

    @property
    def ratio(self):
        return (self.medias + self.files) / self.messages


def dec_hook(type: type, obj: "Any"):
    if type is Path:
        return Path(obj)
    err = f"type {type} is not implemented"
    raise NotImplementedError(err)



__all__ = (
    "JSON_ENC",
    "add_misc_args",
    "add_opts_args",
    "encode_json_str",
    "format_duration",
    "parse_proxy",
    "round_size",
    "unpack_default",
    "wrap_async",
)

from .aiohelper import wrap_async
from .helper import (
    JSON_ENC,
    add_misc_args,
    add_opts_args,
    encode_json_str,
    format_duration,
    parse_proxy,
    round_size,
    unpack_default,
)



TYPE_CHECKING = False
if TYPE_CHECKING:
    from urllib.parse import ParseResult


def create(url: "ParseResult"):
    match url.scheme.lower():
        case "mysqlx":
            from .mysqlx import MySQLXSession as arc

        case "sqlite" | "sqlite3":
            from .custom import CustomSession

            return CustomSession(url.path.strip("/"))

        case "postgresql":
            from .pg import PSQLSession as arc

        case Never:
            err = f"unknown database scheme: {Never}"
            raise ValueError(err)
    return arc(
        user=url.username or "",
        password=url.password or "",
        host=url.hostname or "127.0.0.1",
        port=url.port or 0,
        schema=url.path.strip("/"),
    )



from collections.abc import AsyncIterator
from typing import Any, Literal, overload

from telethon import TelegramClient
from telethon.hints import Entity
from telethon.network import MTProtoSender
from telethon.tl.custom import Message
from telethon.tl.custom.file import File
from telethon.tl.types import TypeInputFileLocation

from tg_util.src.types import EntityStats, FileAttribute

ENTITIES: dict[str, Entity]
SENDERS: dict[int, MTProtoSender]

@overload
async def resolve_entity(
    client: TelegramClient,
    entity: Any,
    with_stats: Literal[True],
) -> tuple[Entity, EntityStats]: ...
@overload
async def resolve_entity(
    client: TelegramClient,
    entity: Any,
    with_stats: bool = False,
) -> Entity: ...
async def get_file_hash(
    client: TelegramClient,
    dc_id: int | None,
    location: TypeInputFileLocation,
) -> bytes:
    """known raises: LocationInvalidError"""

def parse_hashtags(message: Message) -> list[str]: ...
def parse_entity(entity: Any) -> tuple[str, str, str, int]:
    """return: class, title, username, id"""

def get_file_attr(file: File) -> FileAttribute: ...
async def get_entity_stats(client: TelegramClient, entity: Entity) -> EntityStats: ...
def iter_messages(
    client: TelegramClient,
    entity: Entity,
    ids: int | None = None,
    max_id: int = 0,
    min_id: int = 0,
    wait_time: float | None = None,
    reverse: bool = False,
    with_reply: bool = True,
) -> AsyncIterator[tuple[Message, int | None]]: ...



import os
import sys
from pathlib import Path

from telethon.sessions import SQLiteSession

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any


class CustomSession(SQLiteSession):
    _closed: bool

    def __init__(self, name: str) -> None:
        self.__orig_session = Path(sys.argv[0]).with_name(name).with_suffix(".session")
        if not self.__orig_session.exists():
            SQLiteSession(str(self.__orig_session)).close()

        if sys.platform == "win32":
            root = Path("R:\\")
        else:
            root = Path("/tmp")
        self.__temp_session = root / f"{name}_{os.getpid()}.session"
        with (
            open(self.__temp_session, "wb", 0) as fo,
            open(self.__orig_session, "rb", 0) as fi,
        ):
            fo.writelines(fi)
        super().__init__(str(self.__temp_session))
        self._closed = False

    def __repr__(self) -> str:
        return "<%s: %s>" % (self.__class__.__name__, self.filename)

    def __enter__(self):
        return self

    def __exit__(self, *exc: "Any"):
        self.close()

    def close(self):
        if not self._closed:
            super().close()
            self.__orig_session.write_bytes(self.__temp_session.read_bytes())
            with (
                open(self.__orig_session, "wb", 0) as fo,
                open(self.__temp_session, "rb", 0) as fi,
            ):
                fo.writelines(fi)
            self.delete()
            self._closed = True



import logging

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TextIO

logFormat = logging.Formatter(
    r"%(asctime)s: %(module)s.%(funcName)s: %(levelname)s: %(message)s"
)
logHandler: "logging.StreamHandler[TextIO]"


def default_handler():
    global logHandler
    try:
        return logHandler
    except NameError:
        logHandler = logging.StreamHandler()
        logHandler.setFormatter(logFormat)
        return logHandler


def setup_logging(
    loggers: "Iterable[logging.Logger | str]",
    handler: logging.Handler | None = None,
    debug: bool = False,
):
    if not handler:
        handler = default_handler()
    for logger in loggers:
        if isinstance(logger, str):
            logger = logging.getLogger(logger)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG if debug else logging.INFO)



"""pg session for telethon

changes:
- tables must be created manually
- no version table
"""
# pyright: reportIncompatibleMethodOverride=false

from asyncio.events import AbstractEventLoop, new_event_loop
from asyncio.tasks import run_coroutine_threadsafe
from contextlib import contextmanager
from datetime import datetime, timezone
from threading import Thread

from asyncpg import Connection, Record, connect
from telethon.crypto import AuthKey
from telethon.sessions.memory import MemorySession, _SentFileType
from telethon.sessions.sqlite import CURRENT_VERSION as UPSTREAM_VERSION
from telethon.tl.types import (
    InputDocument,
    InputPhoto,
    PeerChannel,
    PeerChat,
    PeerUser,
    TLObject,
)
from telethon.tl.types.updates import State
from telethon.utils import get_peer_id

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any, Coroutine

CURRENT_VERSION = 7


class PSQLSession(MemorySession):
    _conn: "Connection[Record]"
    _loop: AbstractEventLoop
    _schema: str
    _thread: Thread
    _closed: bool

    def __init__(
        self,
        user: str,
        password: str,
        host: str = "127.0.0.1",
        port: int = 5432,
        schema: str = "telethon",
    ):
        if CURRENT_VERSION < UPSTREAM_VERSION:
            err = (
                f"schema version {CURRENT_VERSION} is lower "
                f"than upstream version {UPSTREAM_VERSION}"
            )
            raise RuntimeError(err)
        super().__init__()
        self.save_entities = True
        self._loop = new_event_loop()
        self._thread = Thread(target=self._start_loop)
        self._thread.start()
        self._schema = schema
        self._closed = False
        self._conn = self._wrap_sync(
            connect(
                user=user,
                password=password,
                host=host,
                port=port,
                database=schema,
            )
        )
        result = self._wrap_sync(self._conn.fetchrow("select * from sessions"))
        if result:
            (
                self._dc_id,
                self._server_address,
                self._port,
                key,
                self._takeout_id,
            ) = result
            self._auth_key = AuthKey(data=key)

    def __repr__(self) -> str:
        return "<%s: %s>" % (self.__class__.__name__, self._schema)

    def __enter__(self):
        return self

    def __exit__(self, *exc: "Any"):
        self.close()

    def clone(self, to_instance=None):
        cloned = super().clone(to_instance)
        cloned.save_entities = self.save_entities
        return cloned

    def set_dc(self, dc_id, server_address, port):
        super().set_dc(dc_id, server_address, port)
        self._update_session_table()
        row = self._wrap_sync(self._conn.fetchrow("select auth_key from sessions"))
        if row and row[0]:
            self._auth_key = AuthKey(data=row[0])
        else:
            self._auth_key = None

    @MemorySession.auth_key.setter
    def auth_key(self, value):
        self._auth_key = value
        self._update_session_table()

    @MemorySession.takeout_id.setter
    def takeout_id(self, value):
        self._takeout_id = value
        self._update_session_table()

    def _update_session_table(self):
        with self._transactions():
            self._wrap_sync(self._conn.execute("truncate sessions"))
            self._wrap_sync(
                self._conn.execute(
                    "insert into sessions values($1, $2, $3, $4, $5)",
                    self._dc_id,
                    self._server_address,
                    self._port,
                    self._auth_key.key if self._auth_key else b"",
                    self._takeout_id,
                )
            )

    def get_update_state(self, entity_id: int):
        row = self._wrap_sync(
            self._conn.fetchrow(
                'select pts, qts, "date", seq from update_state where id = $1',
                entity_id,
            )
        )
        if row:
            pts, qts, date, seq = row
            date = datetime.fromtimestamp(date, tz=timezone.utc)
            return State(pts, qts, date, seq, unread_count=0)

    def set_update_state(self, entity_id: int, state: State):
        assert state.date
        self._wrap_sync(
            self._conn.execute(
                "insert into update_state values ($1, $2, $3, $4, $5) "
                "on conflict (id) do update set pts = $2, qts = $3, "
                '"date" = $4, seq = $5',
                entity_id,
                state.pts,
                state.qts,
                state.date.timestamp(),
                state.seq,
            )
        )

    def get_update_states(self):
        rows = self._wrap_sync(self._conn.fetch("select * from update_state"))
        for row in rows:
            yield (
                row[0],
                State(
                    pts=row[1],
                    qts=row[2],
                    date=datetime.fromtimestamp(row[3], tz=timezone.utc),
                    seq=row[4],
                    unread_count=0,
                ),
            )

    def close(self):
        if not self._closed:
            self._wrap_sync(self._conn.close())
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._thread.join()
            self._closed = True

    def process_entities(self, tlo: TLObject):
        if not self.save_entities:
            return

        rows = self._entities_to_rows(tlo)
        if not rows:
            return

        now = int(datetime.now().timestamp())
        self._wrap_sync(
            self._conn.executemany(
                "insert into entities values ($1, $2, $3, $4, $5, $6) "
                "on conflict (id) do update set hash = $2, username = $3, "
                'phone = $4, name = $5, "date" = $6',
                ((*row, now) for row in rows),
            )
        )

    def get_entity_rows_by_phone(self, phone: int):
        return self._wrap_sync(
            self._conn.fetchrow(
                "select id, hash from entities where phone = $1",
                phone,
            )
        )

    def get_entity_rows_by_username(self, username: str):
        rows = self._wrap_sync(
            self._conn.fetch(
                'select id, hash, "date" from entities where username = $1',
                username,
            )
        )
        if not rows:
            return None
        if len(rows) > 1:
            rows.sort(key=lambda t: t[2] or 0)
            self._wrap_sync(
                self._conn.executemany(
                    "update entities set username = null where id = $1",
                    ((r[0],) for r in rows[:-1]),
                )
            )
        row = rows[-1]
        return row[0], row[1]

    def get_entity_rows_by_name(self, name: str):
        return self._wrap_sync(
            self._conn.fetchrow("select id, hash from entities where name = $1", name)
        )

    def get_entity_rows_by_id(self, id: int, exact: bool = True):
        q: tuple[Any, ...] = ()
        if exact:
            q = ("select id, hash from entities where id = $1", id)
        else:
            q = (
                "select id, hash from entities where id in ($1, $2, $3)",
                get_peer_id(PeerUser(id)),
                get_peer_id(PeerChat(id)),
                get_peer_id(PeerChannel(id)),
            )
        return self._wrap_sync(self._conn.fetchrow(*q))

    def get_file(self, md5_digest: bytes, file_size: int, cls: "Any"):
        if row := self._wrap_sync(
            self._conn.fetchrow(
                "select id, hash from sent_files where "
                "md5_digest = $1 and file_size = $2 and type = $3",
                md5_digest,
                file_size,
                _SentFileType.from_type(cls).value,
            )
        ):
            return cls(row[0], row[1])

    def cache_file(self, md5_digest: bytes, file_size: int, instance: "Any"):
        if not isinstance(instance, (InputDocument, InputPhoto)):
            err = f"Cannot cache {type(instance)} instance"
            raise TypeError(err)
        self._wrap_sync(
            self._conn.execute(
                "insert into sent_files values ($1, $2, $3, $4, $5) "
                "on conflict (md5_digest, file_size, type) do update set "
                "id = $4, hash = $5",
                md5_digest,
                file_size,
                _SentFileType.from_type(type(instance)).value,
                instance.id,
                instance.access_hash,
            )
        )

    def _wrap_sync[T](self, coro: "Coroutine[Any, Any, T]"):
        return run_coroutine_threadsafe(coro, self._loop).result()

    @contextmanager
    def _transactions(self):
        t = self._conn.transaction()
        try:
            self._wrap_sync(t.start())
            yield
            self._wrap_sync(t.commit())
        except BaseException:
            self._wrap_sync(t.rollback())
            raise

    def _start_loop(self):
        self._loop.run_forever()
        self._loop.close()
